// Open GL program for S817 (Linux version)
// The purpose of this demo is to exercise the 817 board and display
// the monochrome bitmap images.
// jpeg_demo.c is a simpler demo that shows how to capture with less code
//
// Copyright Sensoray Company Inc. 2008-2009
// default mode is single channel.
// (Some video cards,systems
//  use a lot of CPU for drawing and this does not
//  represent the 817 performance.  Please use s817cmd demo 
//  instead to evaluate CPU usage.)
// Author: D.A.
//
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <sys/time.h>
#include <GL/glut.h>
#include <unistd.h>
#include "s817.h"
#include "s817f.h"


#define MAX_PATH 260
#define ACHAN_START 1
#define ACHAN_END   16
// window positioning
#define XSHIFT 40
#define YSHIFT (YSEP * 4 + 50 - 20)
#define XSEP   (704/G_scale+20)
#define YSEP   (576/G_scale+20)
// forward definitions 
static int savefile(void *pBuf, int flen, char *fname);
static void FrameGrabberQuit(void);
static void glDisplayImage(int idx, GLubyte *image, int w, int h);
static double current_time(void);
static void Usage(void);
static void get_fw_ver(void);
// Globals 
static int ImgWidth[16], ImgHeight[16];
static GLenum ImgFormat;
static GLubyte *glImage[16] = {NULL, NULL, NULL, NULL,
                               NULL, NULL, NULL, NULL,
                               NULL, NULL, NULL, NULL,
                               NULL, NULL, NULL, NULL};
static float    xzoom, yzoom;
static int      g_stopping =0;
static int      G_scale = 4;  // scale down size
static int      board = 1;  // board index
static int      count[16];
static float    fps[16];
static int      invalidCnt[16];
static int      noLock[16];
static int      lastSize[16];
static int      noCmd[16];
static int      bufferMask[16];
static int      snapshot = 0;
static int      G_test_mode = 0;
static int      G_no_display = 0;
static FILE     *G_fdbg = NULL;
static int      bPAL = 0;
static double last_disp_time[16];

// active channels
// all the channels are enabled, but not all are displayed
// OEM should be aware of maximum bandwidth requirements for PCI express
static int chanEn[16] = {       
    1, 1, 1, 1,
    1, 1, 1, 1,
    1, 1, 1, 1,
    1, 1, 1, 1,
};

static int displayEn[16] = {    
    1, 1, 1, 1,
    1, 0, 0, 0,
    1, 0, 0, 0,
    1, 0, 0, 0,
};

static int osdEn[16] = {        
    0, 0, 0, 0,
    0, 0, 0, 0,
    0, 0, 0, 0,
    0, 0, 0, 0,
};


static int vSwSet[4][4] = { {1, 2, 3, 4},
                            {4, 1, 2, 3},
                            {3, 4, 1, 2},
                            {2, 3, 4, 5}};

static int active_channel = 0; // current active channel (channel number
                               // on board minus 1(zero indiced)) 
static int single_channel = 1; // 0-display ONLY the current channel at 
                               //original scale 
                               //1-display all channels scaled down 

static int G_chan_mask = 0;  // channels which have been opened 

static MODE                     mode[16];
// output video 
static unsigned int             vOutEn = 0; // all turned off

GLubyte rasters[][13] = {
    {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00}, 
    {0x00, 0x00, 0x18, 0x18, 0x00, 0x00, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18}, 
    {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x36, 0x36, 0x36, 0x36}, 
    {0x00, 0x00, 0x00, 0x66, 0x66, 0xff, 0x66, 0x66, 0xff, 0x66, 0x66, 0x00, 0x00}, 
    {0x00, 0x00, 0x18, 0x7e, 0xff, 0x1b, 0x1f, 0x7e, 0xf8, 0xd8, 0xff, 0x7e, 0x18}, 
    {0x00, 0x00, 0x0e, 0x1b, 0xdb, 0x6e, 0x30, 0x18, 0x0c, 0x76, 0xdb, 0xd8, 0x70}, 
    {0x00, 0x00, 0x7f, 0xc6, 0xcf, 0xd8, 0x70, 0x70, 0xd8, 0xcc, 0xcc, 0x6c, 0x38}, 
    {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x1c, 0x0c, 0x0e}, 
    {0x00, 0x00, 0x0c, 0x18, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x18, 0x0c}, 
    {0x00, 0x00, 0x30, 0x18, 0x0c, 0x0c, 0x0c, 0x0c, 0x0c, 0x0c, 0x0c, 0x18, 0x30}, 
    {0x00, 0x00, 0x00, 0x00, 0x99, 0x5a, 0x3c, 0xff, 0x3c, 0x5a, 0x99, 0x00, 0x00}, 
    {0x00, 0x00, 0x00, 0x18, 0x18, 0x18, 0xff, 0xff, 0x18, 0x18, 0x18, 0x00, 0x00}, 
    {0x00, 0x00, 0x30, 0x18, 0x1c, 0x1c, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00}, 
    {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00}, 
    {0x00, 0x00, 0x00, 0x38, 0x38, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00}, 
    {0x00, 0x60, 0x60, 0x30, 0x30, 0x18, 0x18, 0x0c, 0x0c, 0x06, 0x06, 0x03, 0x03}, 
    {0x00, 0x00, 0x3c, 0x66, 0xc3, 0xe3, 0xf3, 0xdb, 0xcf, 0xc7, 0xc3, 0x66, 0x3c}, 
    {0x00, 0x00, 0x7e, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x78, 0x38, 0x18}, 
    {0x00, 0x00, 0xff, 0xc0, 0xc0, 0x60, 0x30, 0x18, 0x0c, 0x06, 0x03, 0xe7, 0x7e}, 
    {0x00, 0x00, 0x7e, 0xe7, 0x03, 0x03, 0x07, 0x7e, 0x07, 0x03, 0x03, 0xe7, 0x7e}, 
    {0x00, 0x00, 0x0c, 0x0c, 0x0c, 0x0c, 0x0c, 0xff, 0xcc, 0x6c, 0x3c, 0x1c, 0x0c}, 
    {0x00, 0x00, 0x7e, 0xe7, 0x03, 0x03, 0x07, 0xfe, 0xc0, 0xc0, 0xc0, 0xc0, 0xff}, 
    {0x00, 0x00, 0x7e, 0xe7, 0xc3, 0xc3, 0xc7, 0xfe, 0xc0, 0xc0, 0xc0, 0xe7, 0x7e}, 
    {0x00, 0x00, 0x30, 0x30, 0x30, 0x30, 0x18, 0x0c, 0x06, 0x03, 0x03, 0x03, 0xff}, 
    {0x00, 0x00, 0x7e, 0xe7, 0xc3, 0xc3, 0xe7, 0x7e, 0xe7, 0xc3, 0xc3, 0xe7, 0x7e}, 
    {0x00, 0x00, 0x7e, 0xe7, 0x03, 0x03, 0x03, 0x7f, 0xe7, 0xc3, 0xc3, 0xe7, 0x7e}, 
    {0x00, 0x00, 0x00, 0x38, 0x38, 0x00, 0x00, 0x38, 0x38, 0x00, 0x00, 0x00, 0x00}, 
    {0x00, 0x00, 0x30, 0x18, 0x1c, 0x1c, 0x00, 0x00, 0x1c, 0x1c, 0x00, 0x00, 0x00}, 
    {0x00, 0x00, 0x06, 0x0c, 0x18, 0x30, 0x60, 0xc0, 0x60, 0x30, 0x18, 0x0c, 0x06}, 
    {0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0x00, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00}, 
    {0x00, 0x00, 0x60, 0x30, 0x18, 0x0c, 0x06, 0x03, 0x06, 0x0c, 0x18, 0x30, 0x60}, 
    {0x00, 0x00, 0x18, 0x00, 0x00, 0x18, 0x18, 0x0c, 0x06, 0x03, 0xc3, 0xc3, 0x7e}, 
    {0x00, 0x00, 0x3f, 0x60, 0xcf, 0xdb, 0xd3, 0xdd, 0xc3, 0x7e, 0x00, 0x00, 0x00}, 
    {0x00, 0x00, 0xc3, 0xc3, 0xc3, 0xc3, 0xff, 0xc3, 0xc3, 0xc3, 0x66, 0x3c, 0x18}, 
    {0x00, 0x00, 0xfe, 0xc7, 0xc3, 0xc3, 0xc7, 0xfe, 0xc7, 0xc3, 0xc3, 0xc7, 0xfe}, 
    {0x00, 0x00, 0x7e, 0xe7, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xe7, 0x7e}, 
    {0x00, 0x00, 0xfc, 0xce, 0xc7, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc7, 0xce, 0xfc}, 
    {0x00, 0x00, 0xff, 0xc0, 0xc0, 0xc0, 0xc0, 0xfc, 0xc0, 0xc0, 0xc0, 0xc0, 0xff}, 
    {0x00, 0x00, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xfc, 0xc0, 0xc0, 0xc0, 0xff}, 
    {0x00, 0x00, 0x7e, 0xe7, 0xc3, 0xc3, 0xcf, 0xc0, 0xc0, 0xc0, 0xc0, 0xe7, 0x7e}, 
    {0x00, 0x00, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xff, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3}, 
    {0x00, 0x00, 0x7e, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x7e}, 
    {0x00, 0x00, 0x7c, 0xee, 0xc6, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06}, 
    {0x00, 0x00, 0xc3, 0xc6, 0xcc, 0xd8, 0xf0, 0xe0, 0xf0, 0xd8, 0xcc, 0xc6, 0xc3}, 
    {0x00, 0x00, 0xff, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0}, 
    {0x00, 0x00, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xdb, 0xff, 0xff, 0xe7, 0xc3}, 
    {0x00, 0x00, 0xc7, 0xc7, 0xcf, 0xcf, 0xdf, 0xdb, 0xfb, 0xf3, 0xf3, 0xe3, 0xe3}, 
    {0x00, 0x00, 0x7e, 0xe7, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xe7, 0x7e}, 
    {0x00, 0x00, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xfe, 0xc7, 0xc3, 0xc3, 0xc7, 0xfe}, 
    {0x00, 0x00, 0x3f, 0x6e, 0xdf, 0xdb, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0x66, 0x3c}, 
    {0x00, 0x00, 0xc3, 0xc6, 0xcc, 0xd8, 0xf0, 0xfe, 0xc7, 0xc3, 0xc3, 0xc7, 0xfe}, 
    {0x00, 0x00, 0x7e, 0xe7, 0x03, 0x03, 0x07, 0x7e, 0xe0, 0xc0, 0xc0, 0xe7, 0x7e}, 
    {0x00, 0x00, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0xff}, 
    {0x00, 0x00, 0x7e, 0xe7, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3}, 
    {0x00, 0x00, 0x18, 0x3c, 0x3c, 0x66, 0x66, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3}, 
    {0x00, 0x00, 0xc3, 0xe7, 0xff, 0xff, 0xdb, 0xdb, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3}, 
    {0x00, 0x00, 0xc3, 0x66, 0x66, 0x3c, 0x3c, 0x18, 0x3c, 0x3c, 0x66, 0x66, 0xc3}, 
    {0x00, 0x00, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x3c, 0x3c, 0x66, 0x66, 0xc3}, 
    {0x00, 0x00, 0xff, 0xc0, 0xc0, 0x60, 0x30, 0x7e, 0x0c, 0x06, 0x03, 0x03, 0xff}, 
    {0x00, 0x00, 0x3c, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x3c}, 
    {0x00, 0x03, 0x03, 0x06, 0x06, 0x0c, 0x0c, 0x18, 0x18, 0x30, 0x30, 0x60, 0x60}, 
    {0x00, 0x00, 0x3c, 0x0c, 0x0c, 0x0c, 0x0c, 0x0c, 0x0c, 0x0c, 0x0c, 0x0c, 0x3c}, 
    {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xc3, 0x66, 0x3c, 0x18}, 
    {0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00}, 
    {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x38, 0x30, 0x70}, 
    {0x00, 0x00, 0x7f, 0xc3, 0xc3, 0x7f, 0x03, 0xc3, 0x7e, 0x00, 0x00, 0x00, 0x00}, 
    {0x00, 0x00, 0xfe, 0xc3, 0xc3, 0xc3, 0xc3, 0xfe, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0}, 
    {0x00, 0x00, 0x7e, 0xc3, 0xc0, 0xc0, 0xc0, 0xc3, 0x7e, 0x00, 0x00, 0x00, 0x00}, 
    {0x00, 0x00, 0x7f, 0xc3, 0xc3, 0xc3, 0xc3, 0x7f, 0x03, 0x03, 0x03, 0x03, 0x03}, 
    {0x00, 0x00, 0x7f, 0xc0, 0xc0, 0xfe, 0xc3, 0xc3, 0x7e, 0x00, 0x00, 0x00, 0x00}, 
    {0x00, 0x00, 0x30, 0x30, 0x30, 0x30, 0x30, 0xfc, 0x30, 0x30, 0x30, 0x33, 0x1e}, 
    {0x7e, 0xc3, 0x03, 0x03, 0x7f, 0xc3, 0xc3, 0xc3, 0x7e, 0x00, 0x00, 0x00, 0x00}, 
    {0x00, 0x00, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xfe, 0xc0, 0xc0, 0xc0, 0xc0}, 
    {0x00, 0x00, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x00, 0x00, 0x18, 0x00}, 
    {0x38, 0x6c, 0x0c, 0x0c, 0x0c, 0x0c, 0x0c, 0x0c, 0x0c, 0x00, 0x00, 0x0c, 0x00}, 
    {0x00, 0x00, 0xc6, 0xcc, 0xf8, 0xf0, 0xd8, 0xcc, 0xc6, 0xc0, 0xc0, 0xc0, 0xc0}, 
    {0x00, 0x00, 0x7e, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x78}, 
    {0x00, 0x00, 0xdb, 0xdb, 0xdb, 0xdb, 0xdb, 0xdb, 0xfe, 0x00, 0x00, 0x00, 0x00}, 
    {0x00, 0x00, 0xc6, 0xc6, 0xc6, 0xc6, 0xc6, 0xc6, 0xfc, 0x00, 0x00, 0x00, 0x00}, 
    {0x00, 0x00, 0x7c, 0xc6, 0xc6, 0xc6, 0xc6, 0xc6, 0x7c, 0x00, 0x00, 0x00, 0x00}, 
    {0xc0, 0xc0, 0xc0, 0xfe, 0xc3, 0xc3, 0xc3, 0xc3, 0xfe, 0x00, 0x00, 0x00, 0x00}, 
    {0x03, 0x03, 0x03, 0x7f, 0xc3, 0xc3, 0xc3, 0xc3, 0x7f, 0x00, 0x00, 0x00, 0x00}, 
    {0x00, 0x00, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xe0, 0xfe, 0x00, 0x00, 0x00, 0x00}, 
    {0x00, 0x00, 0xfe, 0x03, 0x03, 0x7e, 0xc0, 0xc0, 0x7f, 0x00, 0x00, 0x00, 0x00}, 
    {0x00, 0x00, 0x1c, 0x36, 0x30, 0x30, 0x30, 0x30, 0xfc, 0x30, 0x30, 0x30, 0x00}, 
    {0x00, 0x00, 0x7e, 0xc6, 0xc6, 0xc6, 0xc6, 0xc6, 0xc6, 0x00, 0x00, 0x00, 0x00}, 
    {0x00, 0x00, 0x18, 0x3c, 0x3c, 0x66, 0x66, 0xc3, 0xc3, 0x00, 0x00, 0x00, 0x00}, 
    {0x00, 0x00, 0xc3, 0xe7, 0xff, 0xdb, 0xc3, 0xc3, 0xc3, 0x00, 0x00, 0x00, 0x00}, 
    {0x00, 0x00, 0xc3, 0x66, 0x3c, 0x18, 0x3c, 0x66, 0xc3, 0x00, 0x00, 0x00, 0x00}, 
    {0xc0, 0x60, 0x60, 0x30, 0x18, 0x3c, 0x66, 0x66, 0xc3, 0x00, 0x00, 0x00, 0x00}, 
    {0x00, 0x00, 0xff, 0x60, 0x30, 0x18, 0x0c, 0x06, 0xff, 0x00, 0x00, 0x00, 0x00}, 
    {0x00, 0x00, 0x0f, 0x18, 0x18, 0x18, 0x38, 0xf0, 0x38, 0x18, 0x18, 0x18, 0x0f}, 
    {0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18}, 
    {0x00, 0x00, 0xf0, 0x18, 0x18, 0x18, 0x1c, 0x0f, 0x1c, 0x18, 0x18, 0x18, 0xf0}, 
    {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x06, 0x8f, 0xf1, 0x60, 0x00, 0x00, 0x00} 
};

GLuint fontOffset;
void makeRasterFont(void)
{
    GLuint i;
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    fontOffset = glGenLists (128);
    for (i = 32; i < 127; i++) {
        glNewList(i+fontOffset, GL_COMPILE);
        glBitmap(8, 13, 0.0, 2.0, 10.0, 0.0, rasters[i-32]);
        glEndList();
    }
}

void myfontinit(void)
{
    makeRasterFont();
}

void printString(char *s)
{
    glPushAttrib(GL_LIST_BIT);
    glListBase(fontOffset);
    glCallLists(strlen(s), GL_UNSIGNED_BYTE, (GLubyte *) s);
    glPopAttrib ();
}

static void getposition( int idx, int *xpos, int *ypos)
{


    switch (idx) {
    case 0:
    case 1:
    case 2:
    case 3:
        *xpos = XSHIFT + XSEP *idx;
        *ypos = YSHIFT;
        break;
    case 4:
    case 5:
    case 6:
    case 7:
        *xpos = XSHIFT + XSEP *(idx-4);
        *ypos = YSHIFT - YSEP;
        break;
    case 8:
    case 9:
    case 10:
    case 11:

        *xpos = XSHIFT + XSEP *(idx-8);
        *ypos = YSHIFT - YSEP *2;
        break;
    case 12:
    case 13:
    case 14:
    case 15:
        *xpos = XSHIFT + XSEP *(idx-12);
        *ypos = YSHIFT - YSEP *3;
        break;

    default:


        break;
    }
    return;
}

static void print_channel(int idx)
{
    char string[100];
    sprintf(string, "Channel %d (use arrows to change, \'z\' to toggle view)\n", idx + 1);
    glRasterPos2i(20, YSHIFT - 20); 
    printString(string);
    if (!displayEn[idx]) {
        sprintf(string, "BMP Off (press 1 to enable)\n");
        glRasterPos2i(20, YSHIFT - 40); 
        printString(string);
    }
    if (!(mode[idx].cMode.capType & CAP_JPG)) {
        sprintf(string, "JPEG Off (press 2 to enable)\n");
        glRasterPos2i(20, YSHIFT - 60); 
        printString(string);
    }
    glRasterPos2i(0, 0);
    return;
}

static void glDisplayImage(int idx, GLubyte *image, int w, int h)
{
    int xpos;
    int ypos;
    static double last_swap=0;
    double t;

    if (image == NULL)
        return;

    if (G_no_display)
        return;

    glPixelStorei(GL_UNPACK_ROW_LENGTH, w);
    glRasterPos3f(0,0,-1.0f);
    ImgWidth[idx] = w;
    ImgHeight[idx] = h;
    if (single_channel) {
        // only display active channel
        if (idx != active_channel) {
            return;
        }
        glClear( GL_COLOR_BUFFER_BIT );
        print_channel(idx);
        glBitmap(0, 0, 0, 0, 50, (GLfloat) YSEP*4+50 - 120, NULL);
        glPixelZoom(1.0, -1.0);
    } else {
        double x_sf = 1.0;
        double y_sf = 1.0;
        if (w < 640 ) {
            x_sf = 2.0;
        }
        if (h <= 576/2 ) {
            y_sf = 2.0;
        }
        getposition(idx, &xpos, &ypos);
        glPixelZoom((GLfloat) x_sf * xzoom/G_scale, (GLfloat) y_sf *-yzoom/G_scale );
        glBitmap(0, 0, 0, 0, (GLfloat) xpos, (GLfloat) ypos, NULL);
    }

    glDrawPixels(w, h, ImgFormat, GL_UNSIGNED_BYTE, image);
    // it's possible to get many calls before the refresh of the screen 
    // calling glutSwapBuffers too often makes video choppy so wait 
    // at least one video frame before updating .030 is just under 
    // the time for one NTSC frame. 
    t = current_time();
    if (t - last_swap > 0.03) {
        last_swap = current_time();
        glutSwapBuffers();
    }
    return;
}

// glut call back
static void glDisplay( void )
{
    int j;
    int xpos;
    int ypos;
    int bh, bw;
    if (G_no_display)
        return;
    glClear( GL_COLOR_BUFFER_BIT );
    if (single_channel) {
        // only display selected active channel
        j = active_channel;
        glClear( GL_COLOR_BUFFER_BIT );
        print_channel(j);
        if (glImage[j] == NULL) {

            glutSwapBuffers();
            return;
        }
        glBitmap(0, 0, 0, 0, (GLfloat) 50, (GLfloat) YSEP*4+50 - 120, NULL);
        glPixelZoom(1.0, -1.0);
        glDrawPixels(ImgWidth[j], ImgHeight[j], ImgFormat, GL_UNSIGNED_BYTE, glImage[j]);

    } else {
        // display all channels 
        for (j = 0; j < 16; j++) {
            double x_sf = 1.0;
            double y_sf = 1.0;
            if (glImage[j] == NULL) {
                continue;
            }
            glRasterPos3f(0,0,-1.0f);
            getposition(j, &xpos, &ypos);
            glBitmap(0, 0, 0, 0, (GLfloat) xpos, (GLfloat) ypos, NULL);
            if (ImgWidth[j] < 640 ) {
                x_sf = 2;
            }
            if (ImgHeight[j] <= 576/2 ) {
                y_sf = 2;
            }
            glPixelZoom( (GLfloat) x_sf * xzoom/G_scale, (GLfloat) y_sf *-yzoom/G_scale );
            glDrawPixels(ImgWidth[j], ImgHeight[j], ImgFormat, GL_UNSIGNED_BYTE, glImage[j]);
        }

        getposition(active_channel, &xpos, &ypos);
        if (bPAL) {
            bw = 704;
            bh = 576;
        } else {
            bw = 640;
            bh = 480;
        }
        glBegin(GL_LINES);
        glColor3ub( 255, 10, 30);
        glVertex2i(xpos - 2, ypos + 2);
        glVertex2i(xpos + bw/G_scale + 2, ypos + 2);
        glEnd();
        glBegin(GL_LINES);
        glVertex2i(xpos + bw/G_scale + 2, ypos + 2);
        glVertex2i(xpos + bw/G_scale + 2, ypos - bh/G_scale - 2);
        glEnd();
        glBegin(GL_LINES);
        glVertex2i(xpos + bw/G_scale + 2, ypos - bh/G_scale - 2);
        glVertex2i(xpos - 2, ypos - bh/G_scale - 2);
        glEnd();
        glBegin(GL_LINES);
        glVertex2i(xpos - 2, ypos - bh/G_scale - 2);
        glVertex2i(xpos - 2 , ypos + 2);
        glEnd();
    }
    glutSwapBuffers();
    return;
}


static void glReshape( int width, int height )
{
    glViewport( 0, 0, width, height );
    glMatrixMode( GL_PROJECTION );
    glLoadIdentity();
    gluOrtho2D( 0.0, width, 0.0, height);
    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();
}

static void display_new_fr( int channel, int rate, int vsys, int bJPG)
{

    static char *NTSCfps[] = {
        "30fps",
        "24fps",
        "20fps",
        "15fps",
        "12fps",
        "10fps",
        "6fps",
        "3fps",
        "2fps",
    };
    static char *PALfps[] = {
        "25fps",
        "20fps",
        "16.7fps",
        "12.5fps",
        "10fps",
        "8.3ps",
        "5fps",
        "2.5fps",
        "1.7fps",
    };

    if (vsys == TV_PAL ) {
        printf("channel %d: %s fps %s\n", channel,
               bJPG ? "JPG" : "BMP",
               PALfps[rate]);
    } else {
        printf("channel %d: %s fps %s\n", channel,
               bJPG ? "JPG" : "BMP",
               NTSCfps[rate]);
    }
    return;
}

static void process_key(unsigned char key, int channel)
{
    int ecode;
    int iParam = 0;
    VIDEO_MODE *vmode;
    CAPTURE_MODE *cmode;
    OSD_MODE osd;

    switch (key) {
    case 'z':
        single_channel = !single_channel;
        break;
    case 'p':
        // switch to PAL
        vmode = &mode[channel].vMode;
        // check if already set to PAL
        if (vmode->tvFormat == TV_PAL)
            break;
        vmode->tvFormat = TV_PAL;
        ecode = S817_SetVideoMode(board, channel+1, vmode);
        if (ecode != 0) {
            if (G_fdbg) {
                fprintf(G_fdbg, "error setting video mode\n");
            } else printf("error setting video mode\n");

            break;
        }
        break;
    case 'n':
        // switch to NTSC
        vmode = &mode[channel].vMode;
        // check if already set to NTSC
        if (vmode->tvFormat == TV_NTSC)
            break;
        vmode->tvFormat = TV_NTSC;
        ecode = S817_SetVideoMode(board, channel+1, vmode);
        if (ecode != 0) {
            if (G_fdbg)
                fprintf(G_fdbg, "error setting video mode\n");
            else printf("error setting video mode\n");
            break;
        }
        break;
    case 'm':
        // switch to sMall size 1CIF
        vmode = &mode[channel].vMode;
        // check if already set to 1CIF 
        if (vmode->jpgSize == SIZE_1CIF)
            break;
        vmode->jpgSize = SIZE_1CIF;
        ecode = S817_SetVideoMode(board, channel+1, vmode);

        if (ecode != 0) {
            if (G_fdbg)
                fprintf(G_fdbg, "error setting video mode\n");
            else printf("error setting video mode\n");;
            break;
        }
        break;
    case 'f':
        // switch to FULL size 4CIF
        vmode = &mode[channel].vMode;
        // check if already set to 4CIF
        if (vmode->jpgSize == SIZE_4CIF)
            break;
        vmode->jpgSize = SIZE_4CIF;
        ecode = S817_SetVideoMode(board, channel+1, vmode);
        if (ecode != 0) {
            if (G_fdbg)
                fprintf(G_fdbg, "error setting video mode\n");
            else printf("error setting video mode\n");
            break;
        }
        break;
    case 'i':
        // switch to FULL size interpolated 4CIFI
        vmode = &mode[channel].vMode;
        // check if already set to 4CIFI
        if (vmode->jpgSize == SIZE_4CIFI)
            break;
        vmode->jpgSize = SIZE_4CIFI;
        ecode = S817_SetVideoMode(board, channel+1, vmode);
        if (ecode != 0) {
            if (G_fdbg)
                fprintf(G_fdbg, "error setting video mode\n");
            else printf("error setting video mode\n");
            break;
        }
        break;
    case 'w':
        // switch to size 2CIF
        vmode = &mode[channel].vMode;
        // check if already set to 2CIF
        if (vmode->jpgSize == SIZE_2CIF)
            break;
        vmode->jpgSize = SIZE_2CIF;
        ecode = S817_SetVideoMode(board, channel+1, vmode);
        if (ecode != 0) {
            if (G_fdbg)
                fprintf(G_fdbg, "error setting video mode\n");
            else printf("error setting video mode\n");
            break;
        }
        break;
    case '1':
        // toggle BMP capture
        cmode = &mode[channel].cMode;
        if (cmode->capType & CAP_BMP ) {
            cmode->capType &= ~CAP_BMP;
        } else {
            cmode->capType |= CAP_BMP;
        }
        ecode = S817_SetCaptureMode(board, channel+1, cmode);
        if (ecode != 0) {
            if (G_fdbg)
                fprintf(G_fdbg,
                        "error setting capture mode %d, err %d\n", 
                        cmode->capType,
                        ecode);
            else printf("error setting capture mode %d, err %d\n", 
                        cmode->capType,
                        ecode);

            break;
        } else {
            int bON = (cmode->capType & CAP_BMP) ? 1 : 0;
            if (!G_fdbg) {
                printf("channel %d: turning %s BMP capture(display)\n",
                       channel+1,
                       bON ? "on" : "off");
            }
            displayEn[channel] = bON;
            if (!bON) {
                glImage[channel] = NULL;
            }
        }
        break;
    case '2':
        // toggle JPG capture
        cmode = &mode[channel].cMode;
        if (cmode->capType & CAP_JPG) {
            cmode->capType &= ~CAP_JPG;
        } else {
            cmode->capType |= CAP_JPG;
        }
        ecode = S817_SetCaptureMode(board, channel+1, cmode);
        if (ecode != 0) {
            if (!G_fdbg)
                printf("error setting capture mode %d\n", 
                       cmode->capType);
            else fprintf(G_fdbg, "error setting capture mode %d\n", 
                         cmode->capType);
            break;
        } else {
            int bON = (cmode->capType & CAP_JPG) ? 1 : 0;
            if (!G_fdbg)
                printf("channel %d: turning %s JPG capture\n",
                       channel+1,
                       bON ? "on" : "off");
        }           
        break;
    case '[':
        // decrease BMP frame rate
        vmode = &mode[channel].vMode;
        cmode = &mode[channel].cMode;
        if (cmode->frameRateBmp == FR_1_15) {
            return;
        }
        cmode->frameRateBmp++;
        ecode = S817_SetCaptureMode(board, channel+1, cmode);
        if (ecode != 0) {
            if (!G_fdbg) 
                printf("error setting capture mode %d\n", 
                       cmode->capType);
            else fprintf(G_fdbg,"error setting capture mode %d\n", 
                         cmode->capType);
            break;
        } else {
            if (!G_fdbg)
                display_new_fr(channel+1, cmode->frameRateBmp, vmode->tvFormat, 0);
        }           

        break;
    case ']':
        // increase BMP frame rate
        vmode = &mode[channel].vMode;
        cmode = &mode[channel].cMode;
        if (cmode->frameRateBmp == FR_FULL) {
            return;
        }
        cmode->frameRateBmp--;
        ecode = S817_SetCaptureMode(board, channel+1, cmode);
        if (ecode != 0) {
            if (!G_fdbg) {
                printf("error setting capture mode %d\n", 
                       cmode->capType);
            } else {
                fprintf(G_fdbg, "error set capture mode %d\n", 
                        cmode->capType);
            }
            break;
        } else {
            if (!G_fdbg) {
                display_new_fr(channel+1, cmode->frameRateBmp, vmode->tvFormat, 0);
            }
        }
        break;
    case '.':
        // increase JPG frame rate
        vmode = &mode[channel].vMode;
        cmode = &mode[channel].cMode;
        if (cmode->frameRateJpg == FR_FULL) {
            return;
        }
        cmode->frameRateJpg--;
        ecode = S817_SetCaptureMode(board, channel + 1, cmode);
        if (ecode != 0) {
            if (!G_fdbg) {
                printf("error setting capture mode %d\n", cmode->capType);
            } else {
                fprintf(G_fdbg, "error set capture mode %d\n", 
                        cmode->capType);
            }
            break;
        } else {
            if (!G_fdbg) {
                display_new_fr(channel+1, cmode->frameRateJpg, vmode->tvFormat, 1);
            }
        }
        break;
    case ',':
        // decrease JPG frame rate
        vmode = &mode[channel].vMode;
        cmode = &mode[channel].cMode;
        if (cmode->frameRateJpg == FR_1_15) {
            return;
        }
        cmode->frameRateJpg++;
        ecode = S817_SetCaptureMode(board, channel+1, cmode);
        if (ecode != 0) {
            if (!G_fdbg) {
                printf("error setting capture mode %d\n", 
                       cmode->capType);
            } else {
                fprintf(G_fdbg, "error set capture mode %d\n", 
                        cmode->capType);
            }
            break;
        } else {
            if (!G_fdbg) {
                display_new_fr(channel+1, 
                               cmode->frameRateJpg, 
                               vmode->tvFormat,
                               1);
            }
        }
        break;
    case ')':
    case '!':
    case '@':
    case '#':
        if (key == ')')
            iParam = 0;
        else if (key == '!')
            iParam = 1;
        else if (key == '@')
            iParam = 2;
        else if (key == '#')
            iParam = 3;
        ecode = S817_SetVideoSwitch(board, &vSwSet[iParam][0]);
        if (ecode != 0) {
            if (!G_fdbg) {
                printf ("Error calling S817_SetVideoSwitch %d\n", ecode);
            }
        } else {
            if (!G_fdbg)
                printf ("Setting video out to %d-%d-%d-%d\n", vSwSet[iParam][0], vSwSet[iParam][1], vSwSet[iParam][2], vSwSet[iParam][3]);
        }
        break;
    case '$':
    case '%':
    case '^':
    case '&':
        if (key == '$')
            iParam = 0;
        else if (key == '%')
            iParam = 1;
        else if (key == '^')
            iParam = 2;
        else if (key == '&')
            iParam = 3;

        iParam = 1 << iParam;
        vOutEn ^= iParam;                               //XOR value with mask: inverts the bit selected in mask
        ecode = S817_SetVideoOut(board, vOutEn);
        if (ecode != 0) {
            if (!G_fdbg)
                printf ("Error calling S817_SetVideoSwitch %d\n", ecode);
        } else {
            if (!G_fdbg)
                printf ("Setting video out enable to %08X\n", vOutEn);
        }
        break;

    case 'j':
    case 'J':
        snapshot = 1;
        break;
    case 'V':
    case 'v':
        memset(&osd, 0, sizeof(osd));
        osdEn[channel] = !osdEn[channel];
        osd.osdOn = osdEn[channel];
        osd.osdBmp = 1;
        osd.transparent = 1;
        osd.positionTop = 1;
        osd.ddmm = 0;
        osd.year2 = 0;
        osd.fraction = 2;
        strncpy(osd.line, "Date: ^d Time: ^t", 80);
        S817_SetOsdMode(board, channel + 1, &osd);
        break;
    case '+':
        vmode = &mode[channel].vMode;
        if (vmode->brightness < 255) 
            vmode->brightness++;
        if ((ecode = S817_SetVideoMode( board, channel + 1, vmode))) {
            if (!G_fdbg)
                printf("err setting video mode\n");
            else
                fprintf(G_fdbg, "err setting video mode\n");
        }
        break;
    case '-':
        vmode = &mode[channel].vMode;
        if (vmode->brightness > 0) 
            vmode->brightness--;
        if ((ecode = S817_SetVideoMode(board, channel + 1, vmode))) {
            if (!G_fdbg)
                printf("err setting video mode\n");
            else
                fprintf(G_fdbg, "err setting video mode\n");
        }
        break;
    case '|':
        G_no_display = !G_no_display;
        printf("debug no display %d\n", G_no_display);
        break;

    case 27:
    case 'q':
    case 'Q':
        g_stopping = 1;
        usleep(1000);
        FrameGrabberQuit();
        if (G_fdbg) {
            fprintf(G_fdbg, "s817 test program ended\n");
            fclose(G_fdbg);
        }
        exit(0);
        break;
    }
}

static void Key( unsigned char key, int x, int y )
{
    (void) x;
    (void) y;
    process_key(key, active_channel);
    if ((key == '1') || (key == 'z') || (key == '2'))
        glutPostRedisplay();
}



static void SpecialKey( int key, int x, int y )
{
    (void) x;
    (void) y;
    switch (key) {
    case GLUT_KEY_UP:
        if (active_channel >= 4) 
            active_channel -= 4;
        break;
    case GLUT_KEY_DOWN:
        if (active_channel < 12) 
            active_channel += 4;
        break;
    case GLUT_KEY_LEFT:
        if (active_channel != 0) 
            active_channel--;
        break;
    case GLUT_KEY_RIGHT:
        if (active_channel < 15) 
            active_channel++;
        break;
    default:
        break;
    }
    glutPostRedisplay();
}


static void init_gl(void)
{
    printf("GL_VERSION = %s\n", (char *) glGetString(GL_VERSION));
    printf("GL_RENDERER = %s\n", (char *) glGetString(GL_RENDERER));
    myfontinit();
    xzoom = 1.0;
    yzoom = 1.0;
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    ImgFormat = GL_LUMINANCE;
    return;
}

// return current time (in seconds)
static double current_time(void)
{
    struct timeval tv;
    struct timezone tz;
    (void) gettimeofday(&tv, &tz);
    return (double) tv.tv_sec + tv.tv_usec / 1000000.0;
}

void idle(void)
{
    char fname[MAX_PATH];
    static int frames = 0;
    int res;
    int aChan;
    int capType;
    int ecode;
    CAP_STAT *capStat; // pointer to capture status struct
    CAP_BUFFER cBuf;
    double cur_time;

    static double tRot0 = -1.0, tRate0 = -1.0;

    if (g_stopping == 1) {
        return;
    }

    res = S817_BlockDone(board, 300, &aChan, &capType);
    if (res != 0) {
        // timeout 
        printf("Timeout err %d \n", res);
        return;
    }
    cur_time = current_time();
    if (capType & CAP_BMP) {
        if ((ecode = S817_GetBuffer(board, aChan, &cBuf, CAP_BMP))) {
            printf ("GetBuffer error %d; aChan %d\n", ecode, aChan);
            return;
        }

        capStat = (CAP_STAT *) cBuf.pStat;
        if (capStat->chanId != aChan)
            printf ("Chan ID error on chan %d: %d\n", aChan, capStat->chanId);

        if (capStat->valid)
            ++count[aChan - 1];
        else
            ++invalidCnt[aChan - 1];

        if (capStat->videopresent == 0)
            ++noLock[aChan - 1];

        lastSize[aChan - 1] = capStat->jpgsize;
        noCmd[aChan - 1] = capStat->reserved[0];
        bufferMask[aChan - 1] = capStat->reserved[1];
        {
            double t = current_time();
            int tot = 0;
            int k;
            for (k = 0; k < 16; k++) {
                if (displayEn[k] || (mode[k].cMode.capType & CAP_JPG))
                    tot++;
            }
            if (tRot0 < 0.0)
                tRot0 = t;

            tRot0 = t;
            frames++;
            if (tRate0 < 0.0)
                tRate0 = t;
            if ((t - tRate0 >= 2.5) && (tot > 0)) {
                GLfloat seconds = (GLfloat) (t - tRate0);
                GLfloat calc_fps = frames / seconds;
                if (!G_fdbg)
                    printf("%d frames in %3.1f seconds = %6.3f fps: %6.3f fps(per channel)\n", frames, seconds, calc_fps, calc_fps/tot);
                else {
                    fprintf(G_fdbg, "%d frms in %3.1f s = %6.3f fps\n", frames, seconds, calc_fps);
                    fflush(G_fdbg);
                }
                tRate0 = t;
                frames = 0;
            }
        }
        // copy the frame to the GL back buffer for display 
        if (displayEn[aChan-1]) {
            glImage[aChan-1] = (unsigned char *) cBuf.pData;
            // don't bother displaying if received new frame before normal refresh rate.
            // which means we(user program) are falling behind in processing frames.
            // 0.02 is 50fps and is reasonable for NTSC and PAL.
            //
            if (cur_time - last_disp_time[aChan-1] > 0.02) {
                glDisplayImage(aChan - 1, glImage[aChan-1], cBuf.width, cBuf.height);
                last_disp_time[aChan-1] = current_time();
            }
        } else {
            glImage[aChan-1] = NULL;
        }
        S817_ReleaseBuffer(board, aChan, cBuf.bufId, CAP_BMP);
    }

    // if snapshot triggered save the jpg 
    if (capType & CAP_JPG) {
        if (!(capType & CAP_BMP))
            frames++;
        if ((ecode = S817_GetBuffer(board, aChan, &cBuf, CAP_JPG))) {
            printf ("GetBuffer error %d; aChan %d\n", ecode, aChan);
            return;
        }
        capStat = (CAP_STAT *) cBuf.pStat;
        if(snapshot && ((aChan - 1) == active_channel)) {
            snapshot = 0;
            if (capStat->jpgsize && capStat->valid &&
                (mode[aChan-1].cMode.capType & CAP_JPG)) {
                sprintf (fname, "snap%02d_%08X", aChan, capStat->tick);
                savefile (cBuf.pData, capStat->jpgsize, fname);
            } else {
                printf("not valid\n");
            }
        }
        S817_ReleaseBuffer(board, aChan, cBuf.bufId, CAP_JPG);
    }
    return;
}



int main(int argc, char **argv )
{
    ECODE           ecode;
    int             i;
    int             aChan;
    MODE            defMode = DEFAULT_MODE;
    int argc_glut = 1;
    int numChannels = SYS_CHAN_PER_BOARD;
    float version;
    if (argc < 2) {
        printf("\n\n\n");
        printf("=========================================\n");
        printf("you should specify the video system and board number\n");
        printf("usage: \"./s817demo system board_num\"\n");
        printf("         where system 0 ==NTSC, 1 ==PAL\n");
        printf("EG:  \"./s817demo 0 1\"\n for NTSC on the first board\n");
        printf("Using defaults board 1 and NTSC\n");
        bPAL = 0;
        board = 1;
        printf("=========================================\n\n");
    } else {
        bPAL = atoi(argv[1]);
    }
    Usage();
#if defined (OSTYPE_WINDOWS)
    _putenv( "__GL_SYNC_TO_VBLANK=1");
#elif defined (OSTYPE_LINUX)
    putenv( "__GL_SYNC_TO_VBLANK=1");
#endif
    if (argc >= 3) {
        board = atoi(argv[2]);
        printf("using board %d\n", board);
        if (board < 1 || board > SYS_BOARDS) {
            printf("invalid board number!\n");
            printf("enter a board between 1 and %d\n", (int) SYS_BOARDS);
            return -1;
        }
        if (argc>3) {
            printf("in test mode logging to 817debug.log\n");
            G_test_mode = 1;
            G_fdbg = fopen("817debug.log", "w+");
            if (G_fdbg) {
                fprintf(G_fdbg, "s817 test program started\n");
                fflush(G_fdbg);
            }
            else printf("open log file failed!\n");
        }
    }
    glutInit( &argc_glut, NULL );
    glutInitWindowPosition( 0, 0 );
    // we display 16 channels in open glut window 
    glutInitWindowSize( XSEP*4 +50, YSEP * 4 + 50);
    glutInitDisplayMode( GLUT_RGB | GLUT_DOUBLE);
    glutCreateWindow(argv[0]);
    init_gl();
    glutReshapeFunc( glReshape );
    glutKeyboardFunc( Key );
    glutSpecialFunc( SpecialKey );
    glutDisplayFunc( glDisplay );

    printf("Version=%s", (char*)glGetString(GL_VERSION));
    version = (float) atof((char*)glGetString(GL_VERSION));
    if ((int)(version *10.0001) < 14) {
        printf("==============================================\n");
        printf("This demo requires an OpenGL video card and/or\n");
        printf("drivers at GL_VERSION 1.4\n");
        printf("Please see the README.  This is not a limitation of the\n");
        printf("817.  The OpenGL video card requirement is only\n");
        printf("for this demo application, which displays video in\n");
        printf("real time\n");
        printf("Please try the other demos described in the README\n");
        printf("and/or upgrade your video card and driver.\n");
        printf("==============================================\n");
        return -1;

    }

    for (i = 0; i < SYS_CHAN_PER_BOARD; i++) {
        count[i] = 0;
        fps[i] = 0.0;
        invalidCnt[i] = 0;
        noLock[i] = 0;
        mode[i] = defMode;
        if (bPAL) {
            mode[i].vMode.tvFormat = TV_PAL;
        }
        // custom mode settings 
        mode[i].cMode.frameRateBmp = FR_FULL;
        mode[i].cMode.frameRateJpg = FR_FULL;
    }

    // Open required channels
    printf("opening channels: Note this may take a while, ");
    printf("but only needs done once at the start of the application\n");
    for (aChan = 1; aChan <= SYS_CHAN_PER_BOARD; aChan++) {
        if (chanEn[aChan - 1]) {
            printf("opening channel %d\n",aChan);
            if ((ecode = S817_OpenChannel(board, aChan, &mode[aChan - 1]))) {
                if (ecode == ERR_ILLEGAL_CHAN) {
                    numChannels--;
                    continue;
                }
                printf ("Chan open error %d; aChan %d\n", ecode, aChan);
                goto eXitus;
            } else {
                if ((ecode = S817_SetVideoMode(board, aChan, &mode[aChan - 1].vMode))) {
                    printf ("SetMode error %d chan %d\n", ecode, aChan);
                    goto eXitus;
                }
                G_chan_mask += (1<<(aChan-1));
            }
        }
    }
    printf("done opening channels.\n");
    printf("IMPORTANT: Frame rate may drop(while displaying images)\n");
    printf("           if your video card is not\n");
    printf("IMPORTANT: openGL compliant OR your driver does not support\n");
    printf("IMPORTANT: all openGL modes.  This is not the fault of\n");
    printf("IMPORTANT: the 817 hardware\n");

    // set the date and time for the board
    S817_SetDateTime(board);

    // Start capture on selected chan's 
    for (aChan = 1; aChan <= SYS_CHAN_PER_BOARD; aChan++) {
        if (chanEn[aChan - 1]) {
            mode[aChan - 1].cMode.capMode = CAP_MODE_CONT;
            mode[aChan - 1].cMode.capType = (aChan == 1) ? CAP_JPG : 0;
            if (displayEn[aChan-1]) {
                mode[aChan - 1].cMode.capType |= CAP_BMP;
            }
            // board, channel, &cMode
            ecode = S817_SetCaptureMode(board, aChan, &mode[aChan - 1].cMode);
            if (ecode && ecode != ERR_ILLEGAL_CHAN) {
                printf ("Capture error %d; channel %d\n", ecode, aChan);
                goto eXitus;
            }
        }
    }
    // get firmware version 
    get_fw_ver();
    glutIdleFunc(idle);
    glutMainLoop();
eXitus:
    FrameGrabberQuit();
    if (G_fdbg) {
        fprintf(G_fdbg, "s817 test program ended\n");
        fclose(G_fdbg);
    }
    return 0;
}

// close down the frame grabber cleanly
static void FrameGrabberQuit(void) 
{
    int i, ecode;
    g_stopping = 1;
    usleep(2000);
    for (i = 0; i < SYS_CHAN_PER_BOARD; i++) {
        if (chanEn[i]) {
            mode[i].cMode.capType = CAP_OFF;
            if ((ecode = S817_SetCaptureMode(board, i + 1, &mode[i].cMode))) {
                printf ("Error %d calling S817_SetCaptureMode\n", ecode);
            }
            if ((ecode = S817_CloseChannel(board, i + 1))) {
                if (ecode != ERR_ILLEGAL_CHAN)
                    printf ("Error %d calling S817_CloseChannel\n", ecode);
            }

        }
    }
    return;
}

// save JPEG to file
static int savefile (void *pBuf, int flen, char *fname)
{
    FILE *fin;
    int ecode = 0;
    char str[MAX_PATH];
    int rc;
    sprintf (str, "%s.jpg", fname);
    printf("savefile %s, len:%d\n",(char *)str, flen);
    fin = fopen( str, "wb+");
    if (fin != NULL) {
        rc = fwrite( pBuf, 1, flen, fin);
        if (rc <= 0) {
            printf("write file error\n");
            ecode = -1;
        }
        fclose(fin);
    } else {
        printf("error\n");
        ecode = -1;
    }
    return ecode;
}


static void Usage(void)
{
    printf("===================================================\n");
    printf("usage: s817demo system(0-NTSC, 1-PAL) board_number(1-4)\n");
    printf("This demo demonstrates some of the capabilities of the 817\n");
    printf("It does not necessarily demonstrate how to configure the 817 for\n");
    printf("your specific application.  Please see manual.\n");
    printf("DISPLAY is monochrome, but snapshots are in color\n");
    printf("Not all Channels are displayed, but snapshots may be taken of\n");
    printf("any channel\n");
    printf("IMPORTANT: Frame rate may drop(while displaying images)\n");
    printf("           if your video card is not\n");
    printf("IMPORTANT: openGL compliant OR your driver does not support\n");
    printf("IMPORTANT: all openGL modes.  This is not the fault of\n");
    printf("IMPORTANT: the 817 hardware\n");
    printf("---------------------------------------------------\n");
    printf("Keys:\n");
    printf("           q  Quit\n");
    printf("     Up/Down  change selected channel\n");
    printf("  Left/Right  change selected channel\n");
    printf("           +  increase brightness\n");
    printf("           -  decrease brightness\n");
    printf("           j  take JPG snapshot of selected channel\n");
    printf("           v  toggle OSD oVerlay on/off\n");
    printf("           1  toggle BMP capture\n");
    printf("           2  toggle JPG capture\n");
    printf("           [  increase frame rate BMP\n");
    printf("           ]  decrease frame rate BMP\n");
    printf("           ,  decrease frame rate JPG\n");
    printf("           .  increase frame rate JPG\n");
    printf("           n  set NTSC tv standard\n");
    printf("           p  set PAL tv standard\n");
    printf("           m  sMall size (1CIF)\n");
    printf("           f  Full size (4CIF)\n");
    printf("           i  Full size interpolated (4CIFI)\n");
    printf("           w  2 CIF size(short/Wide view)\n");
    printf("           z  toggle showing all channels(scaled down)\n");
    printf("              or current channel unscaled(actual size)\n");
    printf("           v  toggle OSD oVerlay on/off\n");
    printf("     shift-0  set video switch 0\n");
    printf("     shift-1  set video switch 1\n");
    printf("     shift-2  set video switch 2\n");
    printf("     shift-3  set video switch 3\n");
    printf("     shift-4  change video output enable\n");
    printf("     shift-5  change video output enable\n");
    printf("     shift-6  change video output enable\n");
    printf("     shift-7  change video output enable\n");
    printf("         ESC  Exit\n");
    printf("===================================================\n");
    return;
}

static void get_fw_ver(void)
{
    SYS_STATUS sysStat;
    ECODE ecode;
    if ((ecode = S817_GetStatusInfo(board, 0, &sysStat))) {
        printf ("Error %d calling S817_GetStatusInfo\n", ecode);
        return;
    }
    if (G_fdbg) {
        fprintf(G_fdbg, "s817 fw version %u.%u\n", (sysStat.fwVer >> 16 & 0xff), (sysStat.fwVer & 0xff));
        fflush(G_fdbg);
    } else 
        printf("s817 fw version %u.%u\n", (sysStat.fwVer >> 16 & 0xff), (sysStat.fwVer & 0xff));
    return;
}
