// Copyright Sensoray Company 2008
// Middleware:  Customers should not and do not need to change this code.
// Source code provided to support as many Linux versions and targets
// as possible

#include <fcntl.h>
#include <poll.h>
#include <stdio.h>
#include <string.h>
#include <sys/mman.h>
#include <time.h>
#include <assert.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <pthread.h>
#include <stdlib.h>
#include "s817ioctl.h"
#include "s817.h"
#include "s817f.h"

#define MAX_PATH 260


static CAP_BUFFER capbuf[SYS_BOARDS][SYS_CHAN_PER_BOARD][SYS_BUF_PER_CHAN][2];
static int G_fd[SYS_BOARDS] = { -1, -1, -1, -1};
static int copen[SYS_BOARDS][SYS_CHAN_PER_BOARD];


/* do an ioctl and it will convert the error code */
static ECODE do_ioctl(int fd, int request, void *arg)
{
    int rc;
    rc = ioctl(fd, request, arg);
    if (rc == -1) {
        switch (errno) {
        case EINVAL: return ERR_ILLEGAL_CHAN;
        case EFAULT: return ERR_ILLEGAL_BUF;
        default: return ERR_UNKNOWN;
        }
    }
    return rc;
}


static ECODE S817_OpenBoard(int board)
{
    char fname[MAX_PATH];
    int fd;
    int bIdx = board - 1;
    int i,j;

    if (bIdx < 0 || bIdx >= SYS_BOARDS)
        return ERR_ILLEGAL_BOARD;

    sprintf(fname, "/dev/s817_%d", bIdx);
    fd = open(fname, O_RDWR);
    G_fd[bIdx] = fd;        
    if (fd == -1)
        return ERR_SX_OPEN;
    for (i=0;i<SYS_CHAN_PER_BOARD;i++) {
        copen[bIdx][i] = 0;
        for(j=0;j<SYS_BUF_PER_CHAN;j++) {
            capbuf[bIdx][i][j][0].pStat = NULL;
            capbuf[bIdx][i][j][0].pData = NULL;
            capbuf[bIdx][i][j][1].pStat = NULL;
            capbuf[bIdx][i][j][1].pData = NULL;
        }
    }
    return 0;

}

static ECODE S817_CloseBoard(int board)
{
    int bIdx = board - 1;

    if (bIdx < 0 || bIdx >= SYS_BOARDS)
        return ERR_ILLEGAL_BOARD;

    close(G_fd[bIdx]);
    G_fd[bIdx] = -1;
    return 0;
}

struct open_chan_param {
    int board;
    int aChan;
    MODE mode;
    ECODE ecode;
};


static ECODE S817_OpenChannel_inner(int board, int aChan, MODE *pMode);

void *open_chan_func( void *arg )
{
    struct open_chan_param *p = (struct open_chan_param *) arg;
    p->ecode = S817_OpenChannel_inner(p->board, p->aChan, &p->mode);
    return NULL;
}



ECODE S817_OpenChannel(int board, int aChan, MODE *pMode)
{
    int bIdx = board - 1;
    int bChan = aChan - 1;
    pthread_attr_t attr;
    ECODE ecode;
    if (bIdx < 0 || bIdx >= SYS_BOARDS)
        return ERR_ILLEGAL_BOARD;

    if (bChan < 0 || bChan >= SYS_CHAN_PER_BOARD)
        return ERR_ILLEGAL_CHAN;

    if (G_fd[bIdx] == -1) {
        S817_OpenBoard(board);
    }

    pthread_attr_init( &attr );
    pthread_attr_setdetachstate( &attr, PTHREAD_CREATE_DETACHED );

    if (aChan == 1) {
        struct open_chan_param param[4];
        pthread_t id[4];
        int i;
        for (i=0;i<4;i++) {
            param[i].aChan = (i*4) + 1;
            param[i].mode = *pMode;
            param[i].board = board;
            pthread_create( &id[i], NULL, &open_chan_func, &param[i] );
        }
        pthread_join(id[0], NULL);
        pthread_join(id[1], NULL);
        pthread_join(id[2], NULL);
        pthread_join(id[3], NULL);
        ecode = 0;
        for (i=0;i<4;i++) {
            if (param[i].ecode != 0 && param[i].ecode != ERR_ILLEGAL_CHAN) {
                ecode = param[i].ecode;
                break;
            }
        }
    } else {
        ecode = S817_OpenChannel_inner(board, aChan, pMode);
    }
    return ecode;
}

static ECODE S817_OpenChannel_inner(int board, int aChan, MODE *pMode)
{
    int rc;
    int j, k;
    int b_fail = 0;
    struct param_openchan p;
    int bIdx = board - 1;
    int bChan = aChan - 1;
    CAP_BUFFER *pCB;

    // already open
    if (copen[bIdx][bChan])
        return 0;
    p.aChan = aChan;
    p.mode = *pMode;
    rc = do_ioctl( G_fd[bIdx], S817_IOCTL_OPENCHANNEL, &p);
    if (rc == 0) {
        struct param_getbuf p2;
        *pMode = p.mode;
        /* mmap all the buffers */
        p2.aChan = aChan;
        for (k = 0; k < 2 ; k++) {
            for (j=0;j<SYS_BUF_PER_CHAN;j++) {
                p2.cBuf.bufId = j;
                p2.capType = k + 1;
                rc = do_ioctl(G_fd[bIdx], S817_IOCTL_GETBUFFER, &p2);
                if (rc == 0) {
                    long size = SYS_BMP_BUF_SIZE;
                    pCB = &capbuf[bIdx][bChan][j][k];
                    pCB->width = p2.cBuf.width;
                    pCB->height = p2.cBuf.height;
                    pCB->bufId = p2.cBuf.bufId;
#ifdef DEBUG_MMAP
                    printf("pStat %p, 0x%lx sizeof %d\n", p2.cBuf.pStat, (off_t) p2.cBuf.pStat,
                           sizeof(struct param_getbuf));
#endif
                    pCB->pStat = mmap( NULL, sizeof( CAP_STAT), PROT_READ | PROT_WRITE, MAP_SHARED, G_fd[bIdx],(off_t) p2.cBuf.pStat);
#ifdef DEBUG_MMAP
                    printf("mmap result %p\n", p2.cBuf.pStat);
#endif
                    if (pCB->pStat == MAP_FAILED) {
                        b_fail = 1;
#ifdef DEBUG_MMAP
                        printf("mmap failed %d!\n", errno);
#endif
                        break;
                    }
                    if (k == (CAP_JPG - 1)) {
                        size = SYS_JPG_BUF_SIZE;
                    } else if (k == (CAP_BMP - 1)) {
                        size = SYS_BMP_BUF_SIZE;
                    }
                    if (p2.cBuf.pData) {
#ifdef DEBUG_MMAP
                        printf("pData %p, 0x%lx\n", p2.cBuf.pData, (off_t) p2.cBuf.pData);
#endif
                        pCB->pData = mmap(NULL, size, PROT_READ | PROT_WRITE, MAP_SHARED, G_fd[bIdx],(off_t) p2.cBuf.pData);
                        if (pCB->pData == MAP_FAILED) {
#ifdef DEBUG_MMAP
                            printf("mmap failed %d!\n", errno);
#endif
                            b_fail = 1;
                            break;
                        }
                    }
                } else {
                    b_fail = 1;
                    break;
                }
            }
            if (b_fail)
                break;
        }


    }
    if (b_fail) {
        // unmap everything
        for (; k >= 0; --k) {
            for (; j >= 0; --j) {
                pCB = &capbuf[bIdx][bChan][j][k];
                if (pCB->pData) {
                    munmap(pCB->pData, (k == 0) ? SYS_JPG_BUF_SIZE :
                            SYS_BMP_BUF_SIZE);
                    pCB->pData = NULL;
                }
                if (pCB->pStat) {
                    munmap(pCB->pStat, sizeof(CAP_STAT));
                    pCB->pStat = NULL;
                }
            }
        }
        return -1;
    }

    if (rc == 0) {
        copen[bIdx][bChan] = 1;
    }
    return rc;
}

ECODE S817_CloseChannel (int board, int aChan)
{
    int rc;
    struct param_closechan p;
    int bChan = aChan - 1;
    int i, k;
    int chan_open = 0;
    int bIdx = board - 1;

    if (bIdx < 0 || bIdx >= SYS_BOARDS)
        return ERR_ILLEGAL_BOARD;

    if (bChan < 0 || bChan >= SYS_CHAN_PER_BOARD)
        return ERR_ILLEGAL_CHAN;

    if (copen[bIdx][bChan] == 0)
        return ERR_CHAN_NOTOPEN;

    p.aChan = aChan;
    rc = do_ioctl(G_fd[bIdx], S817_IOCTL_CLOSECHANNEL, &p);

    for (k=0;k<2;k++) {
        for (i=0;i<SYS_BUF_PER_CHAN;i++) {
            if (capbuf[bIdx][bChan][i][k].pStat) {
                munmap( capbuf[bIdx][bChan][i][k].pStat, sizeof(CAP_STAT));
                capbuf[bIdx][bChan][i][k].pStat = NULL;
            }
            if (capbuf[bIdx][bChan][i][k].pData) {
                munmap( capbuf[bIdx][bChan][i][k].pData,
                        (k == 0) ? SYS_JPG_BUF_SIZE : SYS_BMP_BUF_SIZE);
                capbuf[bIdx][bChan][i][k].pData = NULL;
            }
        }
    }
    copen[bIdx][bChan] = 0;

    for (i = 0; i < SYS_CHAN_PER_BOARD; i++) {
        if ( copen[bIdx][i] ) {
            chan_open = 1;
            break;
        }
    }

    if (chan_open == 0) {
        S817_CloseBoard(board);
    }
    return rc;
}

ECODE S817_SetVideoMode(int board, int aChan, VIDEO_MODE *pVMode)
{
    int rc;
    struct param_setvidmode p;
    int bIdx = board - 1;

    if (bIdx < 0 || bIdx >= SYS_BOARDS)
        return ERR_ILLEGAL_BOARD;

    p.aChan = aChan;
    p.mode = *pVMode;
    rc = do_ioctl( G_fd[bIdx], S817_IOCTL_SETVIDMODE, &p);
    if(rc == 0) {
        *pVMode = p.mode;
    }
    return rc;
}

ECODE S817_SetCaptureMode (int board, int aChan, CAPTURE_MODE *pCMode)
{
    int rc;
    struct param_setcapmode p;
    int bIdx = board - 1;

    if (bIdx < 0 || bIdx >= SYS_BOARDS)
        return ERR_ILLEGAL_BOARD;

    p.aChan = aChan;
    pCMode->capType &= (CAP_JPG | CAP_BMP);
    p.mode = *pCMode;

    rc = do_ioctl( G_fd[bIdx], S817_IOCTL_SETCAPMODE, &p);
    if(rc == 0) {
        *pCMode = p.mode;
    }

    return rc;
}


ECODE S817_GetStatusInfo (int board, int dsp, SYS_STATUS *sysStat)
{
    int rc;
    struct param_getstatusinfo p;
    int bIdx = board - 1;

    if (bIdx < 0 || bIdx >= SYS_BOARDS)
        return ERR_ILLEGAL_BOARD;

    p.dsp = dsp;
    rc = do_ioctl(G_fd[bIdx], S817_IOCTL_GETSTATUSINFO, &p);
    if (rc == 0) {
        *sysStat = p.sysStat;
    }
    return rc;
}

ECODE S817_GetExtErrInfo (int board, int dsp, int chan, EXTERR *err)
{
    int rc;
    struct param_getexterr p;
    int bIdx = board - 1;
    int i;

    if (bIdx < 0 || bIdx >= SYS_BOARDS)
        return ERR_ILLEGAL_BOARD;

    p.dsp = dsp;
    p.errChan = chan;
    rc = do_ioctl(G_fd[bIdx], S817_IOCTL_GETEXTERRINFO, &p);
    if (rc == 0) {
        for ( i = 0; i < SYS_EXTERR_BUF_SIZE; i++) {
            err[i] = p.exterr[i];
        }
    }
    return rc;
}


ECODE S817_GetBuffer(int board, int aChan, CAP_BUFFER *cBuf, int capType)
{
    int rc;
    struct param_getbuf p;
    int bChan = aChan - 1;
    int bIdx = board - 1;

    if (bIdx < 0 || bIdx >= SYS_BOARDS)
        return ERR_ILLEGAL_BOARD;

    if (capType > 2 || capType <= 0)
        return ERR_ILLEGAL_BUFTYPE;

    if (bChan < 0 || bChan >= SYS_CHAN_PER_BOARD)
        return ERR_ILLEGAL_CHAN;

    if (copen[bIdx][bChan] == 0)
        return ERR_CHAN_NOTOPEN;

    p.capType = capType;
    p.aChan = aChan;
    p.cBuf = *cBuf;
    rc = do_ioctl(G_fd[bIdx], S817_IOCTL_GETBUFFERID, &p);

    if (rc == 0) {
        CAP_STAT *pStat;
        int bufId = p.cBuf.bufId;
        *cBuf = p.cBuf;
        cBuf->pStat = capbuf[bIdx][bChan][bufId][capType-1].pStat;
        cBuf->pData = capbuf[bIdx][bChan][bufId][capType-1].pData;
        assert(capbuf[bIdx][bChan][bufId][capType-1].bufId == bufId);
        cBuf->bufId = bufId;
        cBuf->width = p.cBuf.width;
        cBuf->height = p.cBuf.height;
        pStat = (CAP_STAT *) capbuf[bIdx][bChan][bufId][capType-1].pStat;
        pStat->chanId = aChan;
    }
    return rc;
}

ECODE S817_ReleaseBuffer (int board, int aChan, int bufId, int capType)
{
    struct param_relbuf p;
    int bIdx = board - 1;

    if (bIdx < 0 || bIdx >= SYS_BOARDS)
        return ERR_ILLEGAL_BOARD;

    p.aChan = aChan;
    p.capType = capType;
    p.bufId = bufId;
    return do_ioctl(G_fd[bIdx], S817_IOCTL_RELEASEBUF, &p);
}


int lastChan[SYS_BOARDS];
//channel numbers range in aChan context
#define MAX_CHAN        16
#define MIN_CHAN        1
#define NEXT_CHAN(x)    (++(x) > MAX_CHAN) ? MIN_CHAN : x

//starts checking with the channel next after the last one that returned data.
//dataMask has bits [0..15] set/reset according to JPG data present/absent for channels [1..16]
//dataMask has bits [16..31] set/reset according to BMP data present/absent for channels [1..16]
//it's assumed that dataMask != 0 (guaranteed by the caller).
//Returns first channel number having data (lower 16 bits), data type (higher 16 bits)

static int roundRobin(int board, int dataMask)
{
    int     nextChan = 0;
    int dataType = 0;
    int loopcount;

#ifndef _NO_RR_
    nextChan = lastChan[board]; //lastChan is in aChan context [1..16]
#else
    nextChan = 1;
#endif
    loopcount = 0;
    if (dataMask == 0) {
        printf("invalid data mask\n");
        return -1;
    }
    while (!dataType && loopcount < 30) {
        loopcount++;
        nextChan = NEXT_CHAN (nextChan);
        if (dataMask & (1 << (nextChan - 1)))
            dataType |= CAP_JPG;
        if (dataMask & (1 << (nextChan - 1 + 16)))       
            dataType |= CAP_BMP;
    }
    if (loopcount >= 30)
        printf("problem with round robin\n");
    lastChan[board] = nextChan;
    return dataType << 16 | nextChan;
}


//S817_BlockDone
//capType = [CAP_JPG] | [CAP_BMP]
//timeout in ms. If timeout=0 checks and returns immediately.
//Sets aChan - data available;
// and capType - type of data jpg, bmp, both
//If multiple channels available polls in a round robin fashion.
int S817_BlockDone(int board, int timeout, int *aChan, int *capType)
{
    int rc;
    int bIdx = board - 1;
    int rr;
    struct pollfd polllist;
    struct param_getdatamask p1;

    if (bIdx < 0 || bIdx >= SYS_BOARDS)
        return ERR_ILLEGAL_BOARD;

    //if any of the FIFOs have data we return,
    //otherwise we block until data gets written to a FIFO and event is set
    while (1) {
        rc = do_ioctl(G_fd[bIdx], S817_IOCTL_GETDATAMASK, &p1);
        if (rc != 0)
            return ERR_UNKNOWN;
        if (p1.dm)
            break;
        polllist.fd = G_fd[bIdx];
        polllist.events = POLLIN;
        rc = poll(&polllist, 1, timeout);
        if (rc == 0)
            return ERR_CAPTURE_TIMEOUT;
        else if (rc < 0)
            return ERR_UNKNOWN;

    }


    rr = roundRobin(bIdx, p1.dm);
    if (rr != -1 ) {
        *capType = ((rr >> 16) & 0xffff);
        *aChan = (rr & 0xffff);
        return 0;
    } else
        return ERR_UNKNOWN;
}


ECODE S817_SetVideoSwitch(int board, int *switchOut)
{
    struct param_setvidswitch p;

    int bIdx = board - 1;
    if (bIdx < 0 || bIdx >= SYS_BOARDS)
        return ERR_ILLEGAL_BOARD;

    p.switchOut[0] = switchOut[0];
    p.switchOut[1] = switchOut[1];
    p.switchOut[2] = switchOut[2];
    p.switchOut[3] = switchOut[3];
    return do_ioctl(G_fd[bIdx], S817_IOCTL_SETVIDSWITCH, &p);
}

ECODE S817_SetVideoOut(int board, unsigned int outEn)
{
    struct param_setvidout p;
    int bIdx = board - 1;

    if (bIdx < 0 || bIdx >= SYS_BOARDS)
        return ERR_ILLEGAL_BOARD;

    p.outEn = outEn;
    return do_ioctl(G_fd[bIdx], S817_IOCTL_SETVIDOUT, &p);
}

ECODE S817_SetDateTime(int board)
{
    struct param_settime p;
    int bIdx = board - 1;
    struct tm strTm;
    time_t    t;
    if (bIdx < 0 || bIdx >= SYS_BOARDS)
        return ERR_ILLEGAL_BOARD;

    t = time(NULL);
    strTm = *localtime(&t);
    p.tm_sec = strTm.tm_sec;
    p.tm_min = strTm.tm_min;
    p.tm_hour = strTm.tm_hour;
    p.tm_mday = strTm.tm_mday;
    p.tm_mon = strTm.tm_mon;
    p.tm_year = strTm.tm_year;
    p.tm_wday = strTm.tm_wday;
    p.tm_yday = strTm.tm_yday;
    p.tm_isdst = strTm.tm_isdst;
    return do_ioctl(G_fd[bIdx], S817_IOCTL_SETDATETIME, &p);
}

ECODE S817_SetOsdMode(int board, int aChan, OSD_MODE *osd)
{
    struct param_setosdmode p;
    int bIdx = board - 1;
    if (bIdx < 0 || bIdx >= SYS_BOARDS)
        return ERR_ILLEGAL_BOARD;

    p.osdMode = *osd;
    p.aChan = aChan;
    return do_ioctl(G_fd[bIdx], S817_IOCTL_SETOSDMODE, &p);
}



